#!/usr/bin/python3

# @Author : rtekin
# @date : July 2020

# Test of scn traces of hippo execution

# For a given fiacre model :
  # * compile it in hippo,
  # * run hippo executable in a given delay by writing execution trace in a scn file
  # * compile the fiacre model in tts
  # * play this tts in PLAY tool of TINA
  # * load previously got scn in PLAY to check if the execution trace
  # is playable by the tts model

import sys
import subprocess
import threading
import signal
import shlex
import os
import getopt

# Depends on "gen_periodic.py" script of the "measures/exe" directory
sys.path.append("../../../measures")
import exe.gen_periodic as g

# Modify relative paths, overrides gen_periodic's values
g.CONFIG_PATH = "../config.py" # config script to generate hippo Makefile
g.HIPPO_PATH = "../lib/hippo/" # hippo library

FRACPATH = os.getenv("FRACPATH") # FRACPATH environment variable, must be defined
assert FRACPATH != None, "FRACPATH environment variable is not defined"

class Command(object):
    '''
    Enables to run subprocess commands in a different thread with TIMEOUT option
    Based on jcollado's solution:
    http://stackoverflow.com/questions/1191374/subprocess-with-timeout/4825933#4825933
    '''
    def __init__(self, cmd):
        self.cmd = cmd
        self.process = None

    def run(self, timeout=0, **kwargs):
        def target(**kwargs):
            with open(self.cmd + ".scn", "w") as f: # write stdout in a specific file <name>.scn
                self.process = subprocess.Popen(self.cmd, **kwargs, preexec_fn=os.setsid, stdout=f)
                self.process.communicate()

        thread = threading.Thread(target=target, kwargs=kwargs)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            self.process.kill()
            self.killProcess()
            thread.join()
        return self.process.returncode

    def killProcess(self):
        os.killpg(os.getpgid(self.process.pid), signal.SIGKILL)

def fcr2tts(name):
    """
    Compiles a fiacre model in tts by calling the frac compiler

    * Parameters :
      - name : fiacre model name (can be a file path), as a string

    * Side effects :
      - runs two commands with "subprocess.run" method
    """
    format  = {"name" : name, "frac_path" : FRACPATH}
    cmd = shlex.split("frac -tts -g {name}.fcr {name}.tts".format(**format))
    subprocess.run(cmd)
    cmd = shlex.split("gcc -m64  -O2 -o {name}.tts/{name}.so -shared -fPIC -fpic -I {frac_path}/lib -I {frac_path}/lib/avl {name}.tts/{name}.c ".format(**format))
    subprocess.run(cmd)
    return

def run(name, timeout):
    """
    Runs hippo executable

    * Parameters :
      - name : fiacre model name (can be a file path), as a string
      - timeout : execution timeout (in second), None to disable timeout

    * Side effects :
      - runs a command with "command.run" method. See Command clas definition
      abose for more informations.
    """
    path = "./" + name
    try:
        command = Command(path)
        command.run(timeout = timeout)
    except KeyboardInterrupt as e: # execution can be interrupt by "Ctrl+C"
        command.killProcess()

def launchPlay(name) :
    """
    Lauch play of a given tts and loads a scn trace of a hippo execution.

    * Parameters :
      - name : fiacre model name (can be a file path), as a string

    * Side effects :
      - runs a commands with "subprocess.Popen" method
      - writes in stdin
    """
    cmd = shlex.split("play {}.tts".format(name,name))
    p = subprocess.Popen(cmd, stdin = subprocess.PIPE, encoding = 'utf8')
    p.stdin.write('l {}.scn\n'.format(name))
    p.communicate()
    return

def usage():
    print(sys.argv[0], "-n <name> -u <unit> -t <timeout>")
    print("WARNING : All arguments above, except timeout, are mandatory.")
    print("-n <name> : fiacre model name")
    print("-u <unit> : time unit, in second")
    print("-t <timeout> : timeout of hippo execution (in second), disabled by default. Note : Ctrl+C interruption is handled.")
    return

def main():
    name = None
    unit = None
    timeout = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hn:u:t:", ["help"])
    except getopt.GetoptError as err:
        print(str(err))
        usage()
        sys.exit(1)
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o == "-n":
            name = a
        elif o == "-u":
            unit = float(a)
        elif o == "-t":
            timeout = int(a)
        else:
            assert False, "unhandled option"
    if name == None or unit == None :
        print("You have to set a name and a time unit value")
        usage()
        sys.exit(1)
    if ".fcr" in name:
        name = name[:name.index(".fcr")] # remove ".fcr" file extension

    fcr2tts(name)
    g.fcr2hippo(name)
    g.hippo2exec(name, unit, opt_l="", opt_d="-d=scn") # NEED a "main.c" file
    run(name, timeout) # run hippo executable
    launchPlay(name)
    return

if __name__ == "__main__":
    main()
